/* 
 * flood_i.h
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993  Arizona Board of Regents
 *
 */

#ifndef flood_i_h
#define flood_i_h

/*
 * Right now the only session-specific state are flags for IP pseudo-header
 * fixup and flooding control. I could have used separate fields, but got
 * carried by the confounder style of doing it.
 *
 * Flags has the following options right now:
 *
 *   &1 turns on IPpseudoheader length-fixup for session
 *   &2 turns on flooding for that session.
 *
 * IP pseudo-header fixup is not determined solely by &1. Since only sessions
 * which do flooding make any changes to the message, fixup must be allowed
 * only in sessions that have flooding turned on.
 *
 */

typedef struct {
  long  flags; /* &1 turns on IPpseudoheader length-fixup for session */
               /* &2 turns on flooding for outgoing packets */
} SState;    /* session state */

#define is_flood_session(ss) (((SState*)(ss))->flags & 2)

typedef XObj ActiveId;

/*
 * The protocol must keep track of all messages seen in the near past so
 * as to avoid duplicates to be sent forever. This is done by a maintaining
 * an array of mappings that keep track of message identifiers seen.
 * A periodic process is started each FLOOD_LOG_INTERVAL seconds, clean
 * one of the positions of the array (managed as a circular list) and
 * updates the position for new insertions. Message indentifiers are
 * kept for at least ( (FLOOD_LOG_SZ-1) * FLOOD_LOG_INTERVAL ) seconds
 * before they are discarded.
 */

#define FLOOD_LOG_SZ         2
#define FLOOD_LOG_INTERVAL  10
#define FLOOD_LOG_RATE     100 /* messages per interval (average) */


/*
 * We have to keep track of all interfaces, and right now that is done by
 * a simple static array, so we must set its maximu size here.
 */

#define MAX_IF_COUNT        16

/*
 * Messages meant for flooding must carry an unique identifying stamp
 * to avoid duplicates being continuously flooded. This stamp is made
 * by using (one of) the host address and a sequence number. See the
 * comments below.
 */

typedef struct {
  IPhost host_addr;
  long   seqn;
} flood_stamp_t;

/*
 * The protocol state has mapping for active and passive opens, and the
 * following information:
 *
 * Flags, right now only to indicate if IP pseudo-header correction is
 *        needed. As discussed earlier, this information must be combined
 *        with the flooding flag before an action is taken.
 *
 * Messages stamping: next_stamp carries the stamp that must be used for
 *        the next message to be sent. It must be updated each time it is
 *        used by any session.
 *
 * Recently seen messages: msg_log[] and curr_slot implement a time wheel
 *        with mappings to all stamps seen in the near past.
 *
 * Interfaces: if_list[] and if_count can be used to determine all the
 *        interfaces in this host. if_list[0] is currently used as the
 *        host identifier for flood stamps.
 *
 * Addresses that must be flooded: flooding_addresses contains all
 *        addresses for each flooding is allowed. On session creation
 *        this must be checked to determine the session's behavior.
 */

typedef struct {
    Map	 actMap;
    Map  pasMap;
    long flags;	/* &1 turns on IPpseudoheader length-fixup for protocol */
    flood_stamp_t next_stamp;
    Map  msg_log[FLOOD_LOG_SZ];
    int  curr_slot;
    IPhost if_list[MAX_IF_COUNT];
    int    if_count;
    Map  flooding_addresses;
} PState;    /* protocol state */

#define PS_MSG_LOG(ps)   ((ps)->msg_log)
#define PS_CURR_SLOT(ps) ((ps)->curr_slot)
#define PS_MSG_LOG_IDX(ps,idx) ((ps)->msg_log[idx])
#define PS_CURR_MSG_LOG(ps)  ((ps)->msg_log[(ps)->curr_slot])
#define PS_NEXT_SLOT(ps) (PS_CURR_SLOT(ps)=((PS_CURR_SLOT(ps)+1)%FLOOD_LOG_SZ))

/*
 * Right now, the sequence numbers are just that: a sequence. Once one
 * is used, it gets incremented for the next. In order to avoid a denial
 * of service attack, this must be altered to generate an unique random
 * number for each message. This number must be stored for some time to
 * guarantee that a number is not re-used before all hosts have deleted
 * it from their "flooding_messages" log of recently seen messages.
 */

#define generate_new_stamp(ps)  (((ps)->next_stamp.seqn)++)

typedef XObj PassiveId;

#define	FLOOD_ACTIVE_MAP_SZ  64
#define	FLOOD_PASSIVE_MAP_SZ 16
#define	FLOOD_ADDRS_MAP_SZ 16

#define FLOOD_LEN sizeof(flood_stamp_t)

#endif /* flood_i_h */

