/*
 * ocsum-test.c
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Revision: 1.3 $
 * $Date: 1994/04/17 02:08:01 $
 */
/*
 * This is a test program for internet checksumming functions (the
 * function "ocsum" in particular).  To test a new version of "ocsum",
 * add it to file ocsum-fns.c and include an entry for it to array
 * "ocsum_fns" at the end of that file.  Then compile this file with
 * something like:
 *			gcc -O3 ocsum-test.c
 * (file ocsum-fns.c is included by this file).
 *
 * Specify command-line option "-c" to check for correctness and
 * specify "-t" to measure checksumming throughput.
 *
 * Caveat: don't believe in the throughput numbers.  They are merely
 * there for relative performance comparison.  In particular, cache
 * effects are not tested by this program.
 *
 * This program was started by Hilarie Orman (ho@cs.arizona.edu) and
 * extended by David Mosberger-Tang (davidm@cs.arizona.edu).  Any
 * improvements or comments are welcome.
 */
#include <sys/time.h>
#include <sys/types.h>

#define TARGET_TIME	5
#define MAXLEN	(2*64*1024)	/* some ocsum() cannot check longer segs */

int tracecksum = 0;

typedef struct {
    int sec;
    int usec;
} XTime;

typedef struct {
    u_long 	sum;
    int		odd;
} ckSum_t;


#define UNSIGNED(_n) _n##U


static void
subtime(t1, t2, t3)
    XTime *t1, *t2, *t3;
{
    t3->sec = t2->sec - t1->sec;
    t3->usec = t2->usec - t1->usec;
    if (t3->usec < 0) {
	t3->usec += 1000000;
	t3->sec -= 1;
    }
}

void
xGetTime(where)
    XTime *where;
{
    gettimeofday((struct timeval *)where, 0);
}


/*
 * cksum_helper -- perform the checksum over the data.  's' should point to
 * a ckSum_t with the current state of the checksum.  The data pointer does
 * not have to be aligned.
 */
static void
cksum_helper(char *data, long len, void *s, u_short (*ocsum)())
{
    int oddAddr;
    u_long sum = 0;
    int	save_len;

    save_len = len;
    /*
     * Add first byte if necessary to put 'data' on an even address
     */
    oddAddr = ((u_long)data % 2);
    if (oddAddr) {
#if ENDIAN == LITTLE
	sum = *(u_char *)data++ << 8;
#else
	sum = *(u_char *)data++;
#endif
	len--;
    } /* if */
    /*
     * Sum all bytes two at a time, leaving the last one if there are an
     * odd number of bytes:
     */
    if (len > 1) {
	sum += (*ocsum)((u_short *)data, len / 2);
    } /* if */
    data  += len & ~1;
    /*
     * Add in last byte if there is one
     */
    if (len & 1) {
#if ENDIAN == LITTLE
	sum += *(u_char *)data;
#else
	sum += *(u_char *)data << 8;
#endif
    } /* if */
    sum = (sum & 0xffff) + ((sum >> 16) & 0xffff);
    sum = (sum & 0xffff) + ((sum >> 16) & 0xffff);
    /*
     * Swap bytes in the sum if necessary:
     */
    if (oddAddr ^ ((ckSum_t *)s)->odd) {
	/*
	 * Wrap possible overflow:
	 */
	sum = ((sum >> 8) & 0xff) | ((sum & 0xff) << 8);
    } /* if */
    /*
     * Add sum to old sum and indicate whether an odd or even total number
     * of bytes have been processed
     */
    ((ckSum_t *)s)->odd = ((ckSum_t *)s)->odd ^ (save_len % 2);
    ((ckSum_t *)s)->sum += sum;
} /* cksum_helper */

#include "ocsum-fns.c"

static void
usage(char *pgm_name)
{
    printf("usage: %s [ -c ] [ -t ] [ -v ]\n", pgm_name);
    exit(1);
} /* usage */


int
main(int argc, char **argv)
{
    static long darray[2*1024*1024];
    ckSum_t s;
    int i, j, l, ch, sz, off;
    int correctness = 0;
    int throughput = 0;
    int verbose = 0;
    int nerrors = 0;
    int nrounds;
    u_long ref_sum = 0;
    char *test;
    XTime start, stop, result;
    double delta;
    int offset[]  = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15};
    int size[] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
		    63, 125, 213, 419, 731};
    int len[] = {64, MAXLEN, sizeof(darray)/sizeof(darray[0])};

    while ((ch = getopt(argc, argv, "ctv")) != -1) {
	switch (ch) {
	  case 'c':
	    ++correctness; break;
	  case 't':
	    ++throughput; break;
	  case 'v':
	    ++verbose; break;
	  default:
	    usage(argv[0]);
	} /* switch */
    } /* while */
    if (correctness + throughput == 0) {
	usage(argv[0]);
    } /* if */

    /* initialize with some random data: */
    srand48(0);
    for (i = 0; i < sizeof(darray)/sizeof(darray[0]); ++i) {
	darray[i] = lrand48();
    } /* for */

    if (correctness) {
	printf("*** Checking correctness:\n");
	test = "A";
	for (i = 0; i < sizeof(ocsum_fns)/sizeof(ocsum_fns[0]); ++i) {
	    s.sum = 0;
	    s.odd = 0;
	    cksum_helper((char*) darray, MAXLEN, &s, ocsum_fns[i].fun);
	    if (i == 0) {
		ref_sum = s.sum;
	    } /* if */
	    if (verbose) {
		printf("[test %s]: %s computed checksum %#lx\n",
		       test, ocsum_fns[i].name, s.sum);
	    } /* if */
	    if (s.sum != ref_sum) {
		printf("[test %s]: %s computed %#lx, expected %#lx instead!\n",
		       test, ocsum_fns[i].name, s.sum, ref_sum);
		++nerrors;
	    } /* if */
	} /* for */

	/* same with "odd" set to true: */
	test = "B";
	for (i = 0; i < sizeof(ocsum_fns)/sizeof(ocsum_fns[0]); ++i) {
	    s.sum = 0;
	    s.odd = 1;
	    cksum_helper((char*) darray, MAXLEN, &s, ocsum_fns[i].fun);
	    if (i == 0) {
		ref_sum = s.sum;
	    } /* if */
	    if (verbose) {
		printf("[test %s]: %s computed checksum %#lx\n",
		       test, ocsum_fns[i].name, s.sum);
	    } /* if */
	    if (s.sum != ref_sum) {
		printf("[test %s]: %s computed %#lx, expected %#lx instead!\n",
		       test, ocsum_fns[i].name, s.sum, ref_sum);
		++nerrors;
	    } /* if */
	} /* for */

	/* generate second set of data: */
	for (i = 0; i < MAXLEN/sizeof(darray[0]); ++i) {
	    darray[i] = lrand48();
	} /* for */

	test = "C";
	for (i = 0; i < sizeof(ocsum_fns)/sizeof(ocsum_fns[0]); ++i) {
	    s.sum = 0;
	    s.odd = 0;
	    cksum_helper((char*) darray, MAXLEN, &s, ocsum_fns[i].fun);
	    if (i == 0) {
		ref_sum = s.sum;
	    } /* if */
	    if (verbose) {
		printf("[test %s]: %s computed checksum %#lx\n",
		       test, ocsum_fns[i].name, s.sum);
	    } /* if */
	    if (s.sum != ref_sum) {
		printf("[test %s]: %s computed %#lx, expected %#lx instead!\n",
		       test, ocsum_fns[i].name, s.sum, ref_sum);
		++nerrors;
	    } /* if */
	} /* for */

	/* check odd sizes: */
	test = "D";
	for (sz = 0; sz < sizeof(size)/sizeof(size[0]); ++sz) {
	    for (i = 0; i < sizeof(ocsum_fns)/sizeof(ocsum_fns[0]); ++i) {
		s.sum = 0;
		s.odd = 0;
		cksum_helper((char*) darray, size[sz], &s, ocsum_fns[i].fun);
		if (i == 0) {
		    ref_sum = s.sum;
		} /* if */
		if (verbose) {
		    printf("[test %s-%d]: %s computed checksum %#lx\n",
			   test, size[sz], ocsum_fns[i].name, s.sum);
		} /* if */
		if (s.sum != ref_sum) {
		    printf("[test %s-%d]: %s computed %#lx, "
			   "expected %#lx instead!\n",
			   test, size[sz], ocsum_fns[i].name, s.sum, ref_sum);
		    ++nerrors;
		} /* if */
	    } /* for */
	} /* for */

	/* check odd offsetts: */
	test = "E";
	for (off = 0; off < sizeof(offset)/sizeof(offset[0]); ++off) {
	    for (i = 0; i < sizeof(ocsum_fns)/sizeof(ocsum_fns[0]); ++i) {
		s.sum = 0;
		s.odd = 0;
		cksum_helper(((char*) darray) + offset[off], 513, &s,
			     ocsum_fns[i].fun);
		if (i == 0) {
		    ref_sum = s.sum;
		} /* if */
		if (verbose) {
		    printf("[test %s-%d]: %s computed checksum %#lx\n",
			   test, offset[off], ocsum_fns[i].name, s.sum);
		} /* if */
		if (s.sum != ref_sum) {
		    printf("[test %s-%d]: %s computed %#lx, "
			   "expected %#lx instead!\n", test, offset[off],
			   ocsum_fns[i].name, s.sum, ref_sum);
		    ++nerrors;
		} /* if */
	    } /* for */
	} /* for */

	/* check short unaligned segment: */
	test = "F";
	for (i = 0; i < sizeof(ocsum_fns)/sizeof(ocsum_fns[0]); ++i) {
	    s.sum = 0;
	    s.odd = 0;
	    cksum_helper(((char*) darray) + 1, 3, &s, ocsum_fns[i].fun);
	    if (i == 0) {
		ref_sum = s.sum;
	    } /* if */
	    if (verbose) {
		printf("[test %s]: %s computed checksum %#lx\n",
		       test, ocsum_fns[i].name, s.sum);
	    } /* if */
	    if (s.sum != ref_sum) {
		printf("[test %s]: %s computed %#lx, expected %#lx instead!\n",
		       test, ocsum_fns[i].name, s.sum, ref_sum);
		++nerrors;
	    } /* if */
	} /* for */

	/* check medium sized unaligned segment: */
	test = "G";
	for (i = 0; i < sizeof(ocsum_fns)/sizeof(ocsum_fns[0]); ++i) {
	    s.sum = 0;
	    s.odd = 0;
	    cksum_helper(((char*) darray) + 1, 251, &s, ocsum_fns[i].fun);
	    if (i == 0) {
		ref_sum = s.sum;
	    } /* if */
	    if (verbose) {
		printf("[test %s]: %s computed checksum %#lx\n",
		       test, ocsum_fns[i].name, s.sum);
	    } /* if */
	    if (s.sum != ref_sum) {
		printf("[test %s]: %s computed %#lx, expected %#lx instead!\n",
		       test, ocsum_fns[i].name, s.sum, ref_sum);
		++nerrors;
	    } /* if */
	} /* for */

	if (nerrors) {
	    printf("Found %d errors during testing!\n", nerrors);
	} else {
	    printf("Testing completed successfully!\n");
	} /* if */
    } /* if */

    if (throughput) {
	printf("*** Measuring throughput:\n");
	for (l = 0; l < sizeof(len)/sizeof(len[0]); ++l) {
	    printf("**** Checksumming segment of length %d bytes:\n", len[l]);
	    for (i = 0; i < sizeof(ocsum_fns)/sizeof(ocsum_fns[0]); ++i) {
		nrounds = 0;
		do {
		    if (!nrounds) {
			nrounds = 1;
		    } else {
			nrounds *= 2;
		    } /* if */
		    xGetTime(&start);
		    for (j = 0; j < nrounds; ++j) {
			s.sum = 0; s.odd = 0;
			cksum_helper((char*) darray, len[l], &s,
				     ocsum_fns[i].fun);
		    } /* for */
		    xGetTime(&stop);
		    subtime(&start, &stop, &result);
		} while (result.sec == 0 && result.usec < 100*1000);
		delta = (double) result.sec + result.usec / 1000000.0;
		nrounds = (int) ((double) TARGET_TIME / delta * nrounds);
		if (nrounds == 0) {
		    nrounds = 1;
		} /* if */
		xGetTime(&start);
		for (j = 0; j < nrounds; ++j) {
		    s.sum = 0; s.odd = 0;
		    cksum_helper((char*) darray, len[l], &s, ocsum_fns[i].fun);
		} /* for */
		xGetTime(&stop);
		subtime(&start, &stop, &result);
		delta = (double) result.sec + result.usec / 1000000.0;
		printf("%s: %d rounds in %d.%06d sec (%g MB/sec)\n",
		       ocsum_fns[i].name, nrounds, result.sec, result.usec,
		       (double) nrounds * len[l] / delta / (1024.0*1024.0));
	    } /* for */
	} /* for */
    } /* if */
    return nerrors;
} /* main */

			/*** end of ocsum-test.c ***/
