/*     
 * $RCSfile: ipsec-combo.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994,1993,1991,1990  Arizona Board of Regents
 *
 */

/*
 *  The IPSEC header interpreter.
 *
 *  Maps to/from SAID's, for purposes of providing IP security.
 *  Outgoing messages get the remote site's SAID attached and
 *  get encrypted in the key negotiated for that SAID
 *
 *  Incoming messages get decrypted according to the key associated
 *  with the SAID
 *
 *  IP will hand us the full datagram; we must map 
 *        the destination addr, the spi  --> key, alg
 *  then authenticate portions of the ip hdr and all of the datagram text
 *  we'll have to get the spi info out of a key manager
 *
 *  Must keep track of spi's for demuxing to proper session using
 *  next_protocol and spi list?  Yes, and must create new sessions
 *  for new spi-lists and do openDone upward.  This gives server
 *  a chance to get the attributes from each spi and decide whether
 *  or not to accept it.
 *  
 *  Encrypting routers must be prepared to take in non-enhanced messages
 *  and put them through an enhancing session, and vice versa
 *  
 *  
 *  All local maps use the local spi(s) for lookup; if we start with a
 *  remote spi, we must first use the kma mapping to get the local spi.
 *  
 *  spilists are the lookup key; a list must end with a null element;
 *   for mapVar op's, the length is the number of non-zero elements.
 *  
 *  
 */

#include "xkernel.h"
#include "ip.h"
#include "ip/ip_i.h"
#include "saidmgr.h"
#include "km.h"
#include "ip_auth_hdr.h"
#include "ipsec-combo.h"
#include "ipsec-extras.h"
#include "ipsec-combo-sa_i.h"
#include "ipsec-combo_i.h"
#include "ipsec-combo-ah3.h"  /* generated by usit */
#include "hash.h"
#include "crypt.h"

int traceipsecp;
int tracerfc1825p;

void                    rfc1825_init (XObj);
static XObj             ipsecOpen(XObj, XObj, XObj, Part *);
int                     partType(Part);
xkern_return_t   ipsecDemux(XObj, XObj, Msg *);
static xmsg_handle_t    ipsecPush(XObj, Msg *);
static int              ipsecControlSessn(XObj, int, char *, int);
static int              ipsecControlProtl(XObj, int, char *, int);
static xkern_return_t   ipsecClose(XObj);
static ipsec_auth_hdr	*ipsecah(XObj, Msg *, sa_t *, char*, int,
				 ipsec_auth_hdr *);
static spi_t		*spi_overflow(spi_t *, int);
static xkern_return_t	esp_simple_decrypt(Msg *, sa_t *, ISPState *, XObj, char *, int);
static void		ipsecCreateSessn ( XObj );
static long		ivPop( void *, char *, long, void * );
static long		discard( void *, char *, long, void * );
static long		ipsecGetEsp( void *, char *, long, void * );
static void		ivPush( void *, char *, long, void * );
static void		copypush( void *, char *, long, void * );
static long             copypop(void *, char *, long,  void *);
static int		translate_alg(XAlgorithm);
static sa_t		*sa_lookup( spi_t *, ISPState *);
static xmsg_handle_t    esp_simple_encrypt(Msg *, spi_t, char *, ISPState *,
					   char *, int);
static char            *new_iv(int);
static spi_t           *part2flatspilist(Part *);
static xkern_return_t   sa_full_lookup(XObj, IPhost *, spi_t, KMA_MAPPING *);
static XObj             newspisessn(XObj, spi_t *, int, ISSState *);
static int 		calculate_hdr_and_padding_length(XObj,int);
void                    ip_auth_hdr_store( void *, char *, long, void *);
long                    ip_auth_hdr_load( void *, char *, long, void *);
void			ip_esp_hdr_short_store(void *, char *, long, void * );
long			ip_esp_hdr_short_load(void *, char *, long,  void *);
static xkern_return_t   ipsecPassiveOpen(ISPState *, spi_t, int,
					 KMA_MAPPING *, sa_t **,  XObj,
					 ISSState*,  IPheader *, XObj *);
static int reverselist(spi_t *);

/*
The processing of incoming messages is as follows:
loop:	IP extracts header
	if next protocol is IP-AH or IP-ESP, then IP calls IPSEC 
	  to process IP-AH or IP-ESP
	  the message is transformed to:
	    nil -- reject entire message ... exit loop and exit IP
	    valid AH -- ?? could accept message as is for some modes,
			or check spi to see if ESP is next
	    valid ESP mode 1 -- exit loop
	    valid ESP mode 2 -- go to top of loop, message should now start
	else exit loop
end loop
*/

#define XNEW(p, type) { p = ( type * )xMalloc(sizeof(type));\
                        bzero((char *)p, sizeof(type)); }

#define XNEWC(p, type, count) { p = ( type * )xMalloc((count) * sizeof(type));\
                        bzero((char *)p, (count) * sizeof(type)); }

/*
 *  translate_alg
 *
 *  We need a way to map algorithm to down vector index
 *  Could use a rom option.  The rom entries and the protocol
 *  ordering must be carefully planned.
 *
 */
static int translate_alg(i)
     XAlgorithm i;
{
  
  return IPSEC_SPI_LU + i;  /* llp's must be in same order as 
			       indicated in the saidmgr part of the protocol
			       table */
}

/*
    ipsecDemux

    loop over headers  ...
    lookup spi in order to get down session
    pop spi to appropriate down session (will this work?)

    after the loop, we should know the spi sequence and can lookup the
    session; if there isn't one we have to create it now.

    The server can do "getparticipants" and see the spi sequence, do
    any validation it wishes.

    Note that when doing a session lookup for a spi, we must check to
    see if it is a session we are already using under a different name;
    this is because rekeying proposals include changing the spi in
    mid-stream.

    */

xkern_return_t ipsecDemux(s, lls, msg)
     Msg *msg;
     XObj s, lls;
{
  int more_hdrs = 1;
/*  IPpseudoHdr  *ippshdr; */
  ipipsecinfo  *auxinfo = msgGetAttr(msg, 0);
  IPheader     *iphdr = auxinfo->iphdr;
  Msg      ipsec_msg;
  ipsec_auth_hdr ipsecahhdr;
  ipsec_esp_hdr  *ipsecEspHdr;
  sa_t  *rsa;
/*  sa_t  *salist; */
  KMA_MAPPING fsa;
  spi_t *spilist = (spi_t *)xMalloc(MAX_SPIS * sizeof(spi_t));
  spi_t this_spi;
  int spi_index = 0;
  int max_spi_count = MAX_SPIS;
  ISPState *ps = (ISPState *)(s->myprotl->state);
  int ipsec_ah_id  = relProtNumById(protTblGetId("ipsec-ah"), s);
  int ipsec_esp_id = relProtNumById(protTblGetId("ipsec-esp"), s);
  XObj lssn = ERR_XOBJ;
  XObj downs[MAX_SPIS];
  ISSState sstate;
/*  Part ipssn[SPI_PART+1]; */
  int origlen = msgLen(msg);

  bzero((char *)&sstate, sizeof(ISSState));
  msgConstructCopy(&ipsec_msg, msg);
  bzero((char *)spilist, MAX_SPIS*sizeof(spi_t));

  while (more_hdrs) {
    if (iphdr->prot == ipsec_ah_id) {
      spi_t *tspi;
      tspi = (spi_t *)msgPeek(&ipsec_msg, SPI_NET_SIZE+IPSEC_NET_HDR1, FALSE);
      this_spi = ah_peek_spi((char *)tspi);
      if (ipsecPassiveOpen(ps, this_spi,
			   ipsec_ah_id, &fsa, &rsa, s, &sstate,
			   iphdr, downs) != XK_SUCCESS) {
	xFree((char *)spilist); return XK_FAILURE; 
      }
      if (ipsecah(s, &ipsec_msg, rsa, auxinfo->iphdr_net_form,
			       auxinfo->hdr_size, &ipsecahhdr))
	{
	  iphdr->prot = ipsecahhdr.next_hdr;
	  spilist[spi_index++] = ipsecahhdr.spi;
	  xTrace1(rfc1825p, TR_EVENTS, "ipsec adding ah spi 0x%x", ipsecahhdr.spi);
	  if (spi_index == max_spi_count) {
	    spilist = spi_overflow(spilist, max_spi_count);
	    max_spi_count *= 2;
	  }
	}
      else  {   /* AH failure */
	msgDestroy(&ipsec_msg);
	xFree((char *)spilist);
	return XK_FAILURE;
      }
      if ((iphdr->prot != ipsec_ah_id) && (iphdr->prot != ipsec_esp_id)) {
	more_hdrs = 0;
      }
      continue;
    }

    if (iphdr->prot == ipsec_esp_id ) {
      ipsecEspHdr = &ps->topipsechdr;
      msgPop(&ipsec_msg, ipsecGetEsp, ipsecEspHdr, SPI_NET_SIZE, 0);
      if (ipsecPassiveOpen(ps, ipsecEspHdr->spi,
			   ipsec_esp_id, &fsa, &rsa, s, &sstate,
			   iphdr, downs) != XK_SUCCESS) {
	xFree((char *)spilist); return XK_FAILURE; 
      }
      if (fsa.local_mode == IPSEC_ESP_TUNNEL ||
	  fsa.local_mode == IPSEC_ESP_ENCAPS)
	{	
	  char plt;

	  more_hdrs = 0;
	  if (esp_simple_decrypt(&ipsec_msg, rsa, ps, s, &plt, WRD) != XMSG_NULL_HANDLE)
	  { msgDestroy(&ipsec_msg);
	    xFree((char *)ipsecEspHdr);
	    return XK_FAILURE; }
	  if (plt == relProtNumById(protTblGetId("ipinip"), s->myprotl))
	    { }
	  iphdr->prot = plt;
	}
      else {  /* no valid ESP mode for this SPI */
	  msgDestroy(&ipsec_msg);
	  xFree((char *)spilist);
	  return XK_FAILURE;
	}
      if (iphdr->prot != ipsec_ah_id &&
	  iphdr->prot != ipsec_esp_id)
	more_hdrs = 0;
      else more_hdrs = iphdr->prot;
      spilist[spi_index++] = ipsecEspHdr->spi;
      xTrace1(rfc1825p, TR_EVENTS, "ipsec adding esp spi 0x%x", ipsecEspHdr->spi);
      if (spi_index == max_spi_count) {
	spilist = spi_overflow (spilist, max_spi_count);
	max_spi_count *= 2;
      }
      continue;
    }
  }

  msgAssign(msg, &ipsec_msg);
  msgDestroy(&ipsec_msg);

  reverselist(spilist);
  xTrace1(rfc1825p, TR_EVENTS, "ipsecp first spi is 0x%x", *spilist);

  if (mapVarResolve(ps->mapspilist2sessn, spilist, 
		 spi_index * (sizeof(spi_t)), (void *)&lssn)
		 != XK_SUCCESS) {
#ifdef IPSEC_NEVER
    /* if we saved all sa's, we can now create a session referencing
       all the downward sessions 
       
       these will probably be needed for forwarding control op's, etc.

     */
    XObj *ndowns = (XObj *)xMalloc(sizeof(XObj)*spi_index);

    for (j=0; j<spi_index; j++) downs[j] = salist[j].downs; 
#endif /* 0 */
    lssn = newspisessn(s, spilist, spi_index*sizeof(spi_t), &sstate);
    if (lssn == ERR_XOBJ) {
	xTrace3(rfc1825p, TR_ERRORS, 
		"ipsec demux: could not create new session %d %d",
		spilist[0], spilist[1], spi_index);
	return XK_FAILURE;
      }
  }
  else xFree((char *)spilist);

  if (((ISSState *)(lssn->state))->iphack) {
    xTrace1(rfc1825p, TR_FUNCTIONAL_TRACE, "ipsec demux: iphdr fixup %d",
	    origlen-msgLen(msg));
/*    iphdr->dlen -= origlen-msgLen(msg); */
  }
  msgSetAttr(msg, 0, lssn, sizeof(XObj));
  return XK_SUCCESS;
}

static
ipsec_auth_hdr *ipsecah(pself, msg, sa, iphdrnet, hdrsize, ipsecahdr)
     XObj pself;
     Msg *msg;
     sa_t *sa;
     char *iphdrnet;
     int hdrsize;
     ipsec_auth_hdr *ipsecahdr;
{
  ipsec_auth_hdr *ipsecahdr_tmp; 
  spi_t ipsecspi;
  int alg_index;
  xkern_return_t res;
  char *tmp;
  char *tmphash;

  tmp = msgPeek(msg, net_ip_auth_size() + SPI_NET_SIZE + sa->hash_size, FALSE);
  if (!tmp) {
    xTrace2(rfc1825p, TR_SOFT_ERRORS, "ipsec ah: msg too short %d, need %d\n",
	    msgLen(msg), net_ip_auth_size() + SPI_NET_SIZE + sa->hash_size);
    return 0;
  }

/*  msgPop(msg, ip_auth_hdr_spi_peek, &ipsecspi, sizeof(spi_t), 0); */
  if (!sa) {
    xTrace1(rfc1825p, TR_ERRORS, "ipsecah: cannot find session for spi %d\n",
	    ipsecspi);
    return 0;}
  alg_index = translate_alg(sa->alg);  /* local remapping */

  /* this should be part of the state, somewhere */
  XNEW(ipsecahdr_tmp, ipsec_auth_hdr);

  /* need to use spi to lookup the security association */
  /* NB the following does NOT convert the byte order */
  tmphash = xMalloc(sa->hash_size);
  bcopy(tmp+IPSEC_NET_HDR_SIZE, tmphash, sa->hash_size);
  bzero(tmp+IPSEC_NET_HDR_SIZE, sa->hash_size);
  iphdrnet[IPTIME] = 0;
  iphdrnet[CK1] = 0;
  iphdrnet[CK2] = 0;
  msgPush(msg, copypush, iphdrnet, hdrsize, 0);
  msgPush(msg, copypush, tmphash, sa->hash_size, 0);
  xFree((char *)ipsecahdr_tmp);
  xFree(tmphash);
  res = xPop(sa->downs, pself, msg, 0);  /* will md5 know that we are just
					    checking ...
  					the hash isn't part of the message? */
  if (res != XK_FAILURE)  {
    msgPop(msg, discard, 0, hdrsize, 0);
    msgPop(msg, ip_auth_hdr_load, ipsecahdr, IPSEC_NET_HDR_SIZE,
	   0);
    msgPop(msg, discard, 0, sa->hash_size, 0);
    return ipsecahdr;
  }
  return 0;
}


static xkern_return_t esp_simple_decrypt(msg, sa, ps, s, plt, wrd)
     Msg *msg;
     sa_t *sa;
     ISPState *ps;
     XObj s;
     char *plt;
     int wrd;
{
  xkern_return_t res;
  char iv[IPSEC_IV_SIZE];

  if (sa->iv_size > IPSEC_IV_SIZE) {
    xTrace1(rfc1825p, TR_ERRORS, "ipsec decrypt, iv too long %d", sa->iv_size);
    return XK_FAILURE;
  }
  if (sa->iv_size == IPSEC_IV_SIZE/2) {
    int i;

    /* this is from RFC1829 for DES CBC */
    msgPop(msg, ivPop, iv, IPSEC_IV_SIZE/2, 0);
    for (i=0; i<IPSEC_IV_SIZE/2; i++)
      iv[i+IPSEC_IV_SIZE/2] = ~iv[i];
    msgPush(msg, ivPush, iv, IPSEC_IV_SIZE, 0);
  }
  res = xPop(sa->downs, s, msg, 0);  /* handling iv ok? */
  if (res != XK_FAILURE) {

#ifdef IPSEC_NEVER
    char *esp_iv;
    char npad;
    Msg tmsg;

    if (!wrd) 
      msgPop(msg, discard, 0, sa->iv_size, 0);
    msgConstructEmpty(&tmsg);
    if (!wrd) {
      msgChopOff(msg, &tmsg, msgLen(msg)-2);
      msgPop(msg, copypop, &npad, 1, 0);
      msgPop(msg, copypop, plt, 1, 0);
    }
    else 
      msgChopOff(msg, &tmsg, msgLen(msg)-1);
    msgAssign(msg, &tmsg);
    if (wrd) msgPop(msg, copypop, 0, npad, 0);
#endif
    msgPop(msg, copypop, plt, 1, 0);
    return XK_SUCCESS;
  }
    return XK_FAILURE;
}

static xmsg_handle_t esp_simple_encrypt(msg, spi, iv, ps, plt, wrd)
     Msg *msg;
     spi_t spi;
     char *iv;
     ISPState *ps;
     char *plt;
     int  wrd;
{
  sa_t *sa;
  char nv[IPSEC_IV_SIZE];
  xmsg_handle_t ret;

/*  transform must understand IV and payload convention */

  if (mapResolve(ps->mapspi2sa, &spi, &sa) != XK_SUCCESS)
    {
      xTrace1(rfc1825p, TR_SOFT_ERRORS, "ipsec failed to find spi 0x%x", spi);
      return XMSG_ERR_HANDLE;
    }
  msgPush(msg, copypush, plt, 1, 0);
  if (iv) {
    if (sa->iv_size == IPSEC_IV_SIZE/2) {
      int i;

      /* this is from RFC1829 for DES CBC */
      for (i=0; i<IPSEC_IV_SIZE/2; i++)
	nv[i] = iv[i];
      for (i=0; i<IPSEC_IV_SIZE/2; i++)
	nv[i+IPSEC_IV_SIZE/2] = ~iv[i];
      msgPush(msg, ivPush, nv, IPSEC_IV_SIZE, 0);
    }
    else
      msgPush(msg, ivPush, iv, IPSEC_IV_SIZE, 0);
  }
#ifdef IPSEC_NEVER
  long len;

  len = round - (msgLen(msg) % round) - 1;
  npad = len;
  if (!wrd) {
    /* this makes tcrypt compatible with DES CBC transform */
    msgAppend(msg, copypush, zeroes, len, 0, round);
    msgAppend(msg, copypush, &npad, 1, 0, round);
  }
  msgAppend(msg, copypush, plt, 1, 0, round);
#endif
  ret =  xPush(sa->downs, msg);
  if (iv && sa->iv_size == IPSEC_IV_SIZE/2) {
    msgPop(msg, discard, msg, IPSEC_IV_SIZE, 0);
    msgPush(msg, ivPush, iv, IPSEC_IV_SIZE/2, 0);
  }
  return ret;
}

static sa_t *sa_lookup(spi, ps)
     spi_t *spi;
     ISPState *ps;
{
  sa_t *sa = 0;

  if (mapResolve(ps->mapspi2sa, spi, &sa) == XK_SUCCESS)  return sa;
  return 0;
}

static
xmsg_handle_t ipsecPush(s, msg)
     XObj s;
     Msg *msg;
{
  spi_t           *this_spi, *more_spis;
  XIpsecMode      emode;
/*  IPpseudoHdr    *ippshdr; */
  IPheader       *iphdr;   /* the REAL hdr */
/*  ipsec_esp_hdr  *ipsecesp_hdr; */
  sa_t *sa;
  ISSState        *ipsecss = (ISSState *)s->state;
  ISPState        *ps = (ISPState *)s->myprotl->state;
  int		  failed;
  XAlgorithm	  alg;
  ipipsecinfo    *auxinfo;
  int             payload_type;
  char            plt = '\0';
  char *	  tmphash;

  auxinfo = msgGetAttr(msg, 0 /* IPSEC_MSG_ATTR */);
  iphdr = auxinfo->iphdr;
  more_spis = ipsecss->spi_list;
  plt = auxinfo->prot;  /* could be session property */

	
  while ((this_spi = more_spis++, *this_spi)) {
    if (mapResolve(ps->mapspi2sa, this_spi, &sa) == XK_FAILURE)
      {
	xTrace1(rfc1825p, TR_ERRORS, "ipsecPush: cannot find spi %d", *this_spi);
	return XMSG_ERR_HANDLE;
      }
    alg = sa->alg;
    emode = sa->mode;
    switch (sa->mode) {
      case IPSEC_AH: {
	ipsec_auth_hdr  ipsecahhdr;

	xTrace1(rfc1825p, TR_DETAILED, "ipsec push: ah, mode %d\n", emode);

	ipsecahhdr.next_hdr = auxinfo->prot;
	ipsecahhdr.spi = sa->spi;

	ipsecahhdr.length = sa->hash_size/4; /* want # of 32-bit chunks */;
	tmphash=xMalloc(sa->hash_size);
	bzero(tmphash, sa->hash_size);

	xTrace2(rfc1825p, TR_DETAILED, "ipsec push: ah, spi %d hash size %d\n",
		sa->spi,sa->hash_size );

        msgPush(msg,copypush,tmphash,sa->hash_size,0);
	msgPush(msg, ip_auth_hdr_store, &ipsecahhdr,
		net_ip_auth_size( ) + SPI_NET_SIZE , 0);
	msgPush(msg, copypush, auxinfo->iphdr_net_form, auxinfo->hdr_size, 0);
	if ( xPush(sa->downs, msg) != XMSG_ERR_HANDLE)
	  {
	    char *tmp;
	    msgPop(msg, copypop, tmphash, sa->hash_size, 0);
	    msgPop(msg, discard, 0, auxinfo->hdr_size, 0);
	    tmp = msgPeek(msg, IPSEC_NET_HDR_SIZE + sa->hash_size, 0);
	    bcopy (tmphash, tmp+IPSEC_NET_HDR_SIZE, sa->hash_size);
	    auxinfo->prot = relProtNumById(protTblGetId("ipsec-ah"), s->myprotl);
	  }
	else { msgTruncate(msg, 0); failed=1; 
	       xTrace1(rfc1825p, TR_ERRORS, "ipsec ah: crypto failure for spi %x(x)", sa->spi);
	     }
	xFree(tmphash);
      }
      break;

      case IPSEC_ESP_TUNNEL:
      xTrace1(rfc1825p, TR_DETAILED, "ipsec push: esp tunnel, mode %d\n", emode);
	/* we have to attach the full ip header now */
        /* do we want to use a fake source address inside or out? */
	/* for now, we will just use the regular IP hdr */

	plt = (payload_type = relProtNumById(protTblGetId("ipinip"),
					     s->myprotl))==-1?0:payload_type;
	{
	  int tmp = iphdr->dlen;
	  char prot = iphdr->prot;

	  /* as if there were no headers, trailers, etc. */
	  iphdr->dlen = auxinfo->orig_dlen;
	  iphdr->prot = auxinfo->prot;
	  msgPush(msg, ipHdrStore, iphdr, auxinfo->hdr_size, 0);
	  iphdr->dlen = tmp /* + auxinfo->hdr_size */ ; 
	  iphdr->prot = prot;
	}
	/* fall through */

      case IPSEC_ESP_ENCAPS:
	xTrace1(rfc1825p, TR_DETAILED, "ipsec push: esp encaps, mode %d\n", emode);
      /*
	get iv
	encrypt msg, iv
	attach esp hdr
	now ready for ip to attach outer, cleartext hdr, or even for IP-AH

	we need mini-protocols between this and the algs ... they
	attach the transform specific header prior to encryption
	i.e., the padding and payload type
	*/
      {
	char *iv = new_iv(sa->iv_size);
	ipsec_esp_hdr ehdr;

	if (esp_simple_encrypt(msg, *this_spi, iv, ps, &plt, WRD) == XMSG_ERR_HANDLE)
	  {
	    if (iv) xFree(iv);
	    return XMSG_ERR_HANDLE;
	}
	if (iv) {
	  bcopy(iv, (char *)&ehdr.iv, sa->iv_size);
	  xFree(iv);
	}
	ehdr.spi = sa->spi;
	/* last arg is length of data, should always be zero! */
	msgPush(msg, ip_esp_hdr_short_store, &ehdr.spi, SPI_NET_SIZE , 0);
	xIfTrace(rfc1825p, TR_DETAILED) {
	  msgShow(msg);
	}
	/* next protocol should use this protocol as "next hdr" */
	auxinfo->prot = relProtNumById(protTblGetId("ipsec-esp"), s->myprotl);
      }
      break;

    }
  }
  return XMSG_NULL_HANDLE;
}

/* 
   ipsecOpen

      get 3rd participant
       foreach element, get spi, get alg, store is pair list
                        and open down alg with appropriate dst, spi
       createSession, with down sessions
       cross reference ALL alg spi sessions to map
       store in session map
*/
static 
XObj ipsecOpen(self, hlpType, hlpRcv, part)
  XObj self, hlpType, hlpRcv;
  Part *part;
 {
   XObj ns, downs[MAX_SPIS];
   spi_t *spilist, *more_spis, spi;
   Part *spipart, algssn[SPI_PART+1];
   int pl, i = 0;
   ISPState *ps = (ISPState *)self->state;
   IPhost *remotehost, *localhost;

   remotehost  = partPop(part[REMOTE_PART]);
   localhost  = partPop(part[LOCAL_PART]);
   if ((pl=partLen(part)) == SPI_PART+1) spipart = &part[SPI_PART];
   else spipart = &part[SPI_PART-1];
   partPop(spipart[0]);  /* remove the type marker */
   spilist = part2flatspilist(spipart);
   more_spis = spilist;
   while (spi = *more_spis++) {
     sa_t *sa;
     sa_t *lsa;
     KMA_MAPPING km_sa;

     XNEW(sa, sa_t);
     XNEW(lsa, sa_t);

     if (sa_full_lookup(self, localhost, spi, &km_sa) == XK_FAILURE) {
       xFree((char*)sa);        xFree((char*)lsa); 
       return ERR_XOBJ;
     }
     /* create a SA for SENDING to the remote host */
     /*    we also need one for RECEIVING from that host, which we
	   will use in demux                       */

     sa->alg = km_sa.remote_algorithm;
     lsa->alg = km_sa.local_algorithm;
     sa->spi = km_sa.remote_said;
     lsa->spi = spi;
     sa->type = km_sa.remote_mode;  /* make up your mind */
     lsa->type = km_sa.local_mode;  /* make up your mind */
     sa->mode = km_sa.remote_mode;
     lsa->mode = km_sa.local_mode;
     sa->iv_size = km_sa.remote_iv_length;
     lsa->iv_size = km_sa.local_iv_length;
     sa->hash_size = km_sa.remote_hash_size;
     lsa->hash_size = km_sa.local_hash_size;
     sa->padding_type = km_sa.remote_padding_type;
     lsa->padding_type = km_sa.local_padding_type;
     partInit(algssn, 3);
     partPush( algssn[REMOTE_PART], remotehost, sizeof(IPhost));
     partPush( algssn[LOCAL_PART], localhost, sizeof(IPhost));
     partPush( algssn[SPI_PART], &sa->spi, sizeof(spi_t));
     partPush( algssn[SPI_PART], remotehost, sizeof(IPhost));
     downs[i++] =
       xOpen(self, self, xGetDown(self, translate_alg(sa->alg)), algssn);
     if ( downs[i-1] == ERR_XOBJ) {
       xFree((char *)sa);
       xFree((char *)spilist);
       return ERR_XOBJ;
     }
     sa->downs = downs[i-1];
     downs[i-1]->up = 0;
     
     /* check for duplicates ... don't open new session if we don't need to */
     /* however, we must check on the list of open spi sessions ... not the
	spi numbers themselves
	*/
     /* This is for sending */
     mapBind(ps->mapspi2sa, &spi, sa);
     /* this is for receiving, we might not have a session yet */
     mapBind(ps->mapspi2sa_in, &spi, lsa);
     xTrace1(rfc1825p, TR_DETAILED, "ipsecOpen: added local spi %d", spi);
   }
   ns = xCreateSessn(ipsecCreateSessn, hlpType, hlpType, self, i, downs);
   ((ISSState *)ns->state)->spi_list = spilist;
   ((ISSState *)ns->state)->iphack = ps->iphack;
   mapVarBind(ps->mapspilist2sessn, spilist, i*sizeof(spi_t), ns);
   return ns;
 }

static void
ipsecCreateSessn ( self )
     XObj self;
{
  XNEW(self->state, ISSState);
  self->push = ipsecPush;
  self->control = ipsecControlSessn;
/*  self->pop = ipsecPop; */
}


void rfc1825_init(self)
  XObj self;
{
  ISPState *ps;

  XNEW(ps, ISPState);
  ps->mapspilist2sessn = mapCreate(IPSEC_MAP_SZ, -1); /* variable size key */
  ps->mapspi2sa = mapCreate(IPSEC_MAP2_SZ, sizeof(spi_t));
  ps->mapspi2sa_in = mapCreate(IPSEC_MAP2_SZ, sizeof(spi_t));
  self->state = ps;
  self->open = ipsecOpen;
  self->demux = ipsecDemux;
  self->close = ipsecClose;
  self->control = ipsecControlProtl;
  {
    long f = CRYPT_SEND_IV | CRYPT_IV_FROM_MSG | CRYPT_ROTATE_BYTE | CRYPT_RFC_1829;

    xTrace1(rfc1825p, TR_GROSS_EVENTS, "IPSEC setting crypt mode 0x%lx", f);
    xControl(xGetDown(self,4),CRYPT_SET_FLAGS,(char *)&f,sizeof(long));
    xControl(xGetDown(self,6),CRYPT_SET_FLAGS,(char *)&f,sizeof(long));
    xControl(xGetDown(self,8),CRYPT_SET_FLAGS,(char *)&f,sizeof(long));
  }
  xTrace0(rfc1825p, TR_GROSS_EVENTS, "IPSEC STD initializes");
}

static xkern_return_t
ipsecClose(self)
     XObj self;
{
  xTrace0(rfc1825p, TR_MAJOR_EVENTS, "ipsecClose %x");

  if (xIsSession(self)) {
    ISSState *ss = (ISSState *)self;
    xFree((char *)ss);
  }
  return XK_SUCCESS;
}

static int
ipsecControlProtl(self, opcode, buf, len)
  XObj self;
  int opcode;
  char *buf;
  int len;
{
  ISPState *ps = self->state;
    switch (opcode){
    case GETOPTPACKET : return 24;  /* guess, SPI hdr + MD5 */

    case IP_PSEUDOHDR:   /* Turn on flag saying to fixup length on IPpseudoheader. */
      ps->iphack = 1;
      return 0;
    }
  return -1;
}

static int calculate_hdr_and_padding_length(self, dlen)
  XObj self;
  int dlen;
{
  	ISSState *ss;
	int hlen =0;
   	ss = (ISSState *) self->state;
	{
		sa_t * sa = NULL;
		spi_t * cspi = ss->spi_list ,*prev;
 
		while(*cspi && (prev = cspi++))
		{
  			xTrace1(rfc1825p, TR_MAJOR_EVENTS, "ipsec length: calculation spi %u",*prev);
	      		sa = sa_lookup(prev, (ISPState *)self->myprotl->state);
      			if (sa) 
			{
  				xTrace4(rfc1825p, TR_MAJOR_EVENTS, 
				"ipsec length: mode %u,  padding %u, hash %u dlen %u",sa->mode,
				sa->padding_type, sa->hash_size,dlen);
				hlen+=sa->hash_size + sa->iv_size + SPI_NET_SIZE;
				switch(sa->mode) {
					case IPSEC_AH: 
						hlen += net_ip_auth_size();
	 					break;
					case IPSEC_ESP_ENCAPS: 
						hlen += ESP_PAYLOAD_TYPE_SZ;
						break;
					case IPSEC_ESP_TUNNEL: 
						hlen += ESP_PAYLOAD_TYPE_SZ + IPHLEN;
						break;
					default:
						xError("Unknown IPSEC mode");
				}
				switch(sa->padding_type) {
					case SA_NO_PADDING:
						break;
					case SA_8BYTE_DES_PADDING:
						{
						   int pad;
						   /* only tested for DES */
						   pad = 8 - (dlen-IPHLEN+2+
							((sa->mode==IPSEC_ESP_TUNNEL)?IPHLEN:0))%8;
						   hlen +=  pad + 1;
						}
						break;
					default:
						xError("Unknown IPSEC padding type");
						break;
				}
			}	
		}
	}
  	xTrace1(rfc1825p, TR_MAJOR_EVENTS, "ipsec length %u",hlen);
	return hlen;
}

static int
ipsecControlSessn(self, opcode, buf, len)
  XObj self;
  int opcode;
  char *buf;
  int len;
{
  ISSState *ss;

    xAssert(xIsSession(self));

    ss = (ISSState *) self->state;
    switch (opcode){
      
    case IPSEC_GET_MODE:
      {
      /* get the last mode, the one to put into the IP header */
      sa_t *sa = 0;
      spi_t *cspi = ss->spi_list, *prev;

      while (*cspi && (prev = cspi++));
      sa = sa_lookup(prev, (ISPState *)self->myprotl->state);
      if (sa) {
	long protnum;
	char cpr;
	/* really have to translate mode to protocol number ... */
	switch(sa->mode) {
	case IPSEC_AH: 
	  protnum = relProtNumById(protTblGetId("ipsec-ah"), self->myprotl);
	  break;
	case IPSEC_ESP_ENCAPS: 
	case IPSEC_ESP_TUNNEL: 
	  protnum = relProtNumById(protTblGetId("ipsec-esp"), self->myprotl);
	  break;
	default: return -1;
	}
	cpr = protnum;
	bcopy((char *)&cpr, buf, len);
	return sizeof(protnum);
      }
    }
      break;

    case IPSEC_GET_HDRLEN:
      if (ss) {
	u_short *dlen = (u_short *)buf;
	int	x;
	x = calculate_hdr_and_padding_length(self,*dlen);
	*dlen = (u_short) x; 
	return x;
      }
      else return -1;
      break;

      case GETOPTPACKET :
	if (ss) {
	return calculate_hdr_and_padding_length(self,0);
	}
	else return -1;

      case GETPARTICIPANTS:
      {
	    Part	p[SPI_PART+1];
	    int		i = 0;
            spi_t       spi, aspi[MAX_SPIS], *pspi;
/*	    static int   ptysec = - PART_TYPE_SECURE; */

            ss = (ISSState *) self->state;
            pspi = ss->spi_list;

	    partInternalize(p, buf);
	    partInit(&p[SPI_PART], 1);
	    for (i=0; i<SPI_PART+1; i++)
	      p[i].len = SPI_PART+1; /* ??? */
	    i = 0;
            while (spi = *pspi++) aspi[i++] = spi;
            while (--i>=0) 
	      partPush(p[SPI_PART], &aspi[i], sizeof(spi_t));
	    partPush(p[SPI_PART], (void *)-PART_TYPE_SECURE, 0);
            return (partExternalize(p, buf, &len) == XK_FAILURE) ?
		  		-1 : len;
	  }
      break;

      case IP_PSEUDOHDR:   /* Turn on flag saying to fixup length on IPpseudoheader. */
#ifdef IPSEC_NEVER
      if (!ss->iphack) {
	xTrace0(rfc1825p, TR_ERRORS, "ipsec control: ip hdr hack");
	int i;
	ss->iphack = 1;
	for (i=0; i<ss->numdown; i++)
	  xControl(xGetDown(self, i), IP_PSEUDOHDR, 0, 0);
      }
#endif
      break;

      case CRYPT_ISPRIVATE:
      return -1;
      break;

      case IPSEC_GETPEERHOST:
      xTrace0(rfc1825p, TR_ERRORS, "ipsec control: host authentication");
      return sizeof(IPhost);  /* ??? */
      break;

    default:
      xTrace1(rfc1825p, TR_SOFT_ERRORS, "ipsec unrecognized opcode %d", opcode);
    }
  return -1;
}


static long		ipsecGetEsp(dst, src, len, arg)
     void *dst;
     char *src;
     long len;
     void *arg;
{
  spi_t tspi;

  bcopy(src, (char *)&tspi, SPI_NET_SIZE);
  *(unsigned int *)dst = ntohl(*(unsigned int *)&tspi);
  return len;
}

static long ivPop(dst, src, len, arg)
     void *dst;
     char *src;
     long len;
     void *arg;
{
  /* get iv in host byte order */
  xTrace0(rfc1825p, TR_ERRORS, "ipsec ivPop does not get host order");
  if (dst != 0) bcopy(src, dst, len);
  else xTrace0(rfc1825p, TR_SOFT_ERROR, "ipsec ivPop with null dst");
  return len;
}

static long discard(dst, src, len, arg)
     void *dst;
     char *src;
     long len;
     void *arg;
{
  return len;
}

static void ivPush(src, dst, len, arg)
     void *src;
     long len;
     char *dst;
     void *arg;
{
  /* get iv in host byte order */
  xTrace2(rfc1825p, TR_ERRORS, "ipsec ivPush %d bytes, first 0x%x", len, 
	  *(char *)src);
  if (dst != 0) bcopy(src, dst, len);
  else xTrace0(rfc1825p, TR_SOFT_ERROR, "ipsec ivPop with null dst");
}

static void copypush(src, dst, len, arg)
     void *src;
     long len;
     char *dst;
     void *arg;
{
  if (dst) bcopy(src, dst, len);
}

long copypop(dst, src, len, arg)
     void *dst;
     long len;
     char *src;
     void *arg;
{
  if (dst) bcopy(src, (char *)dst, len);
  return len;
}


static spi_t *spi_overflow(spi_list, count)
     spi_t *spi_list;
     int count;
{
  spi_t *new_list = (spi_t *)xMalloc(2*count*sizeof(spi_t));
  int i;
  
  bzero((char *)new_list, 2*count*sizeof(spi_t));
  for(i=0; i<count; i++) new_list[i] = spi_list[i];
  return new_list;
}

static char            *new_iv(len)
int len;
{
  static seed = 1000011;
  char *ans, *tans;

  if (len <= 0) return 0;
  ans = xMalloc(len);

  tans=ans;
  while (len--) {
    seed = ((seed + 13) * 4567) + 5;
    *tans++ = (seed>>16) & 0xff;
  }
  return ans;
}


static int
reverselist(spil)
spi_t *spil;
{
  spi_t *cspi = spil, *nspil;
  int i = 0, j=0;

  while (*cspi++) i++;
  --cspi;
  nspil = (spi_t *) xMalloc(sizeof(spi_t)*(i+1));
  while (i--) nspil[j++] = *--cspi;
  nspil[j] = 0;
  while (nspil[++i]) *spil++ = nspil[i];
  *spil = 0;
  xFree((char *)nspil);
  return i*sizeof(spi_t);
}

static spi_t *
part2flatspilist(part)
  Part *part;
{
  spi_t *spil, *spis;
  int snum=0, stot=0, slen = 1;
  Part *cpart;
  void *selm;

  cpart = (Part *)xMalloc(sizeof(Part));
  partInit(cpart, 1);
  XNEWC(spis, spi_t, (part->stack.top + 1));  /* one extra for termn cond */

  while (slen>=0) {
    if ((slen = partStackTopByteLen(part[0])) >= 0)
	{ 
	  selm = partPop(part[0]);
	  partPush(cpart[0], selm, slen);
	  stot += slen; 
	  snum++;
	}
  }
  if (snum>0) XNEWC(spil, spi_t, (snum + 1));  /* one extra for termn cond */
  while (snum-- > 0)
    {
      selm = partPop(cpart[0]);
      spil[snum] = *(int *)selm;
/*      bcopy((char *)selm, (char *)spil[snum], sizeof(spi_t)); */
      partPush(part[0], selm, sizeof(spi_t));
    }
  xFree((char *)cpart);
  return spil;
}  




int
partType(Part part)
{
  int len;

  len = partStackTopByteLen(part);
  if (len>=0) 
    return (int) partPop(part);
  else return 0;
}

xkern_return_t sa_full_lookup(self, host, spi, km_sa)
     XObj self;
     IPhost *host;
     spi_t spi;
     KMA_MAPPING *km_sa;
{
     XObj spilus;
     xkern_return_t retm;
     Part spilu[1];

     bzero((char *)km_sa, sizeof(KMA_MAPPING));
/*     if ( localhost == ANY_HOST ) localhost = find_local_host(); */
     partInit(spilu, 1);
     partPush(spilu[0], &spi, sizeof(spi_t));
     partPush(spilu[0], host, sizeof(IPhost));
     spilus = xOpen(self, self, xGetDown(self, IPSEC_SPI_LU), spilu);
     if (spilus == ERR_XOBJ) {
       xTrace5(rfc1825p, TR_ERRORS,
 	       "ipsec sa lookup: cannot resolve SPI 0x%x host %x.%x.%x.%x",
	       spi, host->a, host->b, host->c, host->d);
       return XK_FAILURE;
     } else
       if (xControl(spilus, KM_RESOLVE,
		    (char *)km_sa, sizeof(KMA_MAPPING)) < 0) {
	 xTrace1(rfc1825p, TR_ERRORS,
		 "ipsec sa lookup: cannot resolve SPI data structure 0x%x",
		  spi);
	 retm = XK_FAILURE;
       }
       else retm = XK_SUCCESS;
     xClose(spilus);
     return retm;
   }

XObj
newspisessn(s, spil, n, sstate)
     XObj s;
     int n;
     spi_t *spil;
     ISSState *sstate;
{
	XObj lssn;
	ISSState *sasessn;

	XNEW(sasessn, ISSState);
	lssn = xCreateSessn(ipsecCreateSessn, s, s, s, 0, 0);
	lssn->state = sasessn;
	sasessn->spi_list = spil;
	sasessn->self = lssn;
	sasessn->iphack = ((ISPState *)(s->state))->iphack;
	/* might want to use the alg down sessions as the key */
	mapVarBind(((ISPState *)(s->myprotl->state))->mapspilist2sessn,
		   spil, n, lssn);
	return lssn;
      }

static xkern_return_t
ipsecPassiveOpen(ps, spi, mode, fsa, remsa, self, sstate, iphdr, downs)
     ISPState *ps;
     spi_t spi;
     int mode;
     KMA_MAPPING *fsa;
     sa_t **remsa;
     XObj self, *downs;
     ISSState *sstate;
     IPheader     *iphdr;
 {
   xkern_return_t ret;
   sa_t *rsa;
   int i = 0;

      xTrace1(rfc1825p, TR_DETAILED,
		 "ipsecPassiveOpen: SPI %u",spi);
      ret = mapResolve(ps->mapspi2sa_in, &spi, &rsa);
      if (ret == XK_FAILURE)  {
	/* we are doing a passive open */

	sa_t *lsa;
	Part algparts[SPI_PART+1];

	XNEW(lsa, sa_t);
	XNEW(rsa, sa_t);
	if (sa_full_lookup(self, &iphdr->dest, spi, fsa) == XK_FAILURE)
	  { xFree((char *)lsa); xFree((char *)rsa); 
	    return XK_FAILURE; }
	if (translate_alg(fsa->local_algorithm)<=IPSEC_SPI_LU)
	  { xFree((char *)lsa); xFree((char *)rsa); 
	    return XK_FAILURE; }
	partInit( algparts, 3);
	partPush( algparts[REMOTE_PART], &iphdr->source, sizeof(IPhost));
	partPush( algparts[LOCAL_PART],  &iphdr->dest, sizeof(IPhost));
	partPush( algparts[SPI_PART], &spi, sizeof(spi_t));
	partPush( algparts[SPI_PART], &iphdr->dest, sizeof(IPhost));
	downs[i++] =
	  xOpen(self, self, xGetDown(self, translate_alg(fsa->local_algorithm)),
		algparts);
	downs[i-1]->up = 0;

	/* this is what this RECEIVER needs */
	rsa->spi = spi;
	rsa->iv_size = fsa->local_iv_length;
     	rsa->hash_size = fsa->local_hash_size;
     	rsa->padding_type = fsa->local_padding_type;
 	rsa->downs = downs[i-1];
	rsa->mode = fsa->local_mode;
	mapBind(ps->mapspi2sa_in, &spi, rsa);

	/* this is what the SENDER will need */
	partInit( algparts, 3);
	partPush( algparts[REMOTE_PART], &iphdr->source, sizeof(IPhost));
	partPush( algparts[LOCAL_PART],  &iphdr->dest, sizeof(IPhost));
	partPush( algparts[SPI_PART], &fsa->remote_said, sizeof(spi_t));
	partPush( algparts[SPI_PART], &iphdr->source, sizeof(IPhost));
	lsa->downs =
	  xOpen(self, self, xGetDown(self, 
				     translate_alg(fsa->remote_algorithm)),
		algparts);
	lsa->downs->up = 0;
	lsa->spi = fsa->remote_said;
	lsa->alg = fsa->remote_algorithm;
	lsa->mode = fsa->remote_mode;
	lsa->iv_size = fsa->remote_iv_length;
     	lsa->hash_size = fsa->remote_hash_size;
     	lsa->padding_type = fsa->remote_padding_type;
	mapBind(ps->mapspi2sa, &spi, lsa);
	xTrace1(rfc1825p, TR_DETAILED, "ipsec session  iv size %d\n", rsa->iv_size);
      }
      else {
	if (!rsa->downs)
	  {
	    Part algparts[SPI_PART+1];

	    xTrace1(rfc1825p, TR_SOFT_ERROR,
		      "ipsecpPassiveOpen: open sessn for SPI 0x%x", spi);

	    partInit( algparts, 3);
	    partPush( algparts[REMOTE_PART], &iphdr->source, sizeof(IPhost));
	    partPush( algparts[LOCAL_PART], &iphdr->dest, sizeof(IPhost));
	    partPush( algparts[SPI_PART], &spi, sizeof(spi_t));
	    partPush( algparts[SPI_PART], &iphdr->dest, sizeof(IPhost));
	    downs[i++] =
	      xOpen(self, self, xGetDown(self, translate_alg(rsa->alg)),
		    algparts);
	    if (downs[i-1] != ERR_XOBJ) {
	      downs[i-1]->up = 0;
	      rsa->downs = downs[i-1];
	    }
	    else
	    {
	      xTrace1(rfc1825p, TR_SOFT_ERROR,
		      "ipsecpPassiveOpen: could not open local sessn for 0x%x", spi);
	      return XK_FAILURE;
	    }
	  }
 	  fsa->local_mode = rsa->mode;
     }
   *remsa = rsa;
   return XK_SUCCESS;
 }
